/*
 * (C) Copyright 2002
 * Sysgo Real-Time Solutions, GmbH <www.elinos.com>
 * Marius Groeger <mgroeger@sysgo.de>
 *
 * (C) Copyright 2002
 * David Mueller, ELSOFT AG, <d.mueller@elsoft.ch>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <s3c2440.h>
#include <video_fb.h>

DECLARE_GLOBAL_DATA_PTR;

#define FCLK_SPEED 1

#if FCLK_SPEED==0       /* Fout = 203MHz, Fin = 12MHz for Audio */
#define M_MDIV  0xC3
#define M_PDIV  0x4
#define M_SDIV  0x1
#elif FCLK_SPEED==1     /* Fout = 399.65MHz */
#define M_MDIV  0x6e
#define M_PDIV  0x3
#define M_SDIV  0x1
#endif

#define USB_CLOCK 1

#if USB_CLOCK==0
#define U_M_MDIV    0xA1
#define U_M_PDIV    0x3
#define U_M_SDIV    0x1
#elif USB_CLOCK==1
#define U_M_MDIV    0x3c
#define U_M_PDIV    0x4
#define U_M_SDIV    0x2
#endif

#define S3C2410_LCDCON1_TFT1BPP    (8<<1)
#define S3C2410_LCDCON1_TFT2BPP    (9<<1)
#define S3C2410_LCDCON1_TFT4BPP    (10<<1)
#define S3C2410_LCDCON1_TFT8BPP    (11<<1)
#define S3C2410_LCDCON1_TFT16BPP   (12<<1)
#define S3C2410_LCDCON1_TFT24BPP   (13<<1)

#define S3C2410_LCDCON1_TFT        (3<<5)
#define S3C2410_LCDCON1_CLKVAL(x)  ((x) << 8)

#define S3C2410_LCDCON2_VBPD(x)     ((x) << 24)
#define S3C2410_LCDCON2_LINEVAL(x)  ((x) << 14)
#define S3C2410_LCDCON2_VFPD(x)     ((x) << 6)
#define S3C2410_LCDCON2_VSPW(x)     ((x) << 0) 

#define S3C2410_LCDCON3_HOZVAL(x)   ((x) << 8) 
#define S3C2410_LCDCON3_HBPD(x)     ((x) << 19)
#define S3C2410_LCDCON3_HFPD(x)     ((x) << 0)

#define S3C2410_LCDCON4_MVAL(x)     ((x) << 8)
#define S3C2410_LCDCON4_HSPW(x)     ((x) << 0)

#define S3C2410_LCDCON5_FRM565      (1<<11)
#define S3C2410_LCDCON5_INVVLINE    (1<<9)
#define S3C2410_LCDCON5_INVVFRAME   (1<<8)
#define S3C2410_LCDCON5_PWREN       (1<<3)
#define S3C2410_LCDCON5_BSWP        (1<<1) 
#define S3C2410_LCDCON5_HWSWP       (1<<0) 


#define HOZVAL_TFT                  (LCD_XSIZE_TFT-1)
#define LINEVAL_TFT                 (LCD_YSIZE_TFT-1)
#define SCR_XSIZE_TFT               (LCD_XSIZE_TFT)
#define SCR_YSIZE_TFT               (LCD_YSIZE_TFT)

#define M5D(n)                      ((n) & 0x1fffff)

static inline void delay (unsigned long loops)
{
    __asm__ volatile ("1:\n"
      "subs %0, %1, #1\n"
      "bne 1b":"=r" (loops):"0" (loops));
}

/*
 * Miscellaneous platform dependent initialisations
 */

int board_init (void)
{
    S3C24X0_CLOCK_POWER * const clk_power = S3C24X0_GetBase_CLOCK_POWER();
    S3C24X0_GPIO * const gpio = S3C24X0_GetBase_GPIO();

    /* to reduce PLL lock time, adjust the LOCKTIME register */
//  clk_power->LOCKTIME = 0xFFFFFF;

    /* configure MPLL */
//  clk_power->MPLLCON = ((M_MDIV << 12) + (M_PDIV << 4) + M_SDIV);

    /* some delay between MPLL and UPLL */
//  delay (4000);

    /* configure UPLL */
//  clk_power->UPLLCON = ((U_M_MDIV << 12) + (U_M_PDIV << 4) + U_M_SDIV);

    /* some delay between MPLL and UPLL */
//  delay (8000);

    /* set up the I/O ports */
    gpio->GPACON = 0x007FFFFF;
    gpio->GPBCON = 0x002a9655;
    gpio->GPBUP = 0x000007FF;
    gpio->GPCCON = 0xAAAAAAAA;
    gpio->GPCUP = 0x0000FFFF;
    gpio->GPDCON = 0xAAAAAAAA;
    gpio->GPDUP = 0x0000FFFF;
    gpio->GPECON = 0xAAAAAAAA;
    gpio->GPEUP = 0x0000FFFF;
    gpio->GPFCON = 0x000055AA;
    gpio->GPFUP = 0x000000FF;
    gpio->GPGCON = 0xFD95FFBA;
    gpio->GPGUP = 0x0000FFFF;
#ifdef CONFIG_SERIAL3
    gpio->GPHCON = 0x002AAAAA;
#else
    gpio->GPHCON = 0x002AFAAA;
#endif
    gpio->GPHUP = 0x000007FF;

    gpio->GPJCON = 0x2AAAAAA;

#if 0
    /* USB Device Part */
    /*GPGCON is reset for USB Device */
    gpio->GPGCON = (gpio->GPGCON & ~(3 << 24)) | (1 << 24); /* Output Mode */
    gpio->GPGUP = gpio->GPGUP | ( 1 << 12);         /* Pull up disable */

    gpio->GPGDAT |= ( 1 << 12) ; 
    gpio->GPGDAT &= ~( 1 << 12) ; 
    udelay(20000);
    gpio->GPGDAT |= ( 1 << 12) ; 
#endif

    /* arch number of ARMFUN_TSM01-Board */
    gd->bd->bi_arch_number = MACH_TYPE_S3C2440;

    /* adress of boot parameters */
    gd->bd->bi_boot_params = 0x30000100;

    icache_enable();
    dcache_enable();

    return 0;
}

int dram_init (void)
{
    gd->bd->bi_dram[0].start = PHYS_SDRAM_1;
    gd->bd->bi_dram[0].size = PHYS_SDRAM_1_SIZE;

    return 0;
}

int bmp_log_init(int nOption)
{
  const char * pLoadCmd = "nand read.jffs2 %s %s %s";
  char cmdBuff[50];
  char *s;
  ulong addr;
 
  if ((s = getenv ("splashimage")) == NULL) return -1;

  sprintf( cmdBuff , pLoadCmd, s, NAND_LOG_BMP_OFFSET, NAND_LOG_BMP_SIZE);

  run_command (cmdBuff, 0);

  return 0;

}

void board_video_init(GraphicDevice *pGD)
{
    S3C24X0_LCD * const lcd = S3C24X0_GetBase_LCD();
    S3C24X0_GPIO * const gpio = S3C24X0_GetBase_GPIO();
    S3C24X0_TIMERS * const timer = S3C24X0_GetBase_TIMERS();
    
    timer->TCFG0 = 24;
    timer->TCFG1 &= ~(0xf);
    
    timer->ch[0].TCMPB = 2500;
    timer->ch[0].TCNTB = 5000;
    
    timer->TCON &= ~(0xf);
    timer->TCON |= 0xa;
    timer->TCON &= ~(0xf);
    timer->TCON |= 0x9;

    //===========Lcd port init==========//
    gpio->GPCUP  = 0xffffffff; // Disable Pull-up register
    gpio->GPCCON = 0xaaaaaaaa; //Initialize VD[7:0],LCDVF[2:0],VM,VFRAME,VLINE,VCLK,LEND
    gpio->GPDUP  = 0xffffffff; // Disable Pull-up register
    gpio->GPDCON = 0xaaaaaaaa; //Initialize VD[23:8]
    
    //===========Lcd init==========//
#if defined CONFIG_8BIT_INDEX
    // TFT LCD panel,8bpp TFT,ENVID=off
    lcd->LCDCON1 = S3C2410_LCDCON1_TFT8BPP |
                   S3C2410_LCDCON1_TFT |
                   S3C2410_LCDCON1_CLKVAL(0x09);
#elif defined CONFIG_16BIT_565RGB
    lcd->LCDCON1 = S3C2410_LCDCON1_TFT16BPP |
                   S3C2410_LCDCON1_TFT |
                   S3C2410_LCDCON1_CLKVAL(0x09);
#elif defined CONFIG_24BIT_888RGB
    lcd->LCDCON1 = S3C2410_LCDCON1_TFT24BPP |
                   S3C2410_LCDCON1_TFT |
                   S3C2410_LCDCON1_CLKVAL(0x09);
#endif
    lcd->LCDCON2 = S3C2410_LCDCON2_VBPD(7) |
                   S3C2410_LCDCON2_LINEVAL(LINEVAL_TFT) |
                   S3C2410_LCDCON2_VFPD(6) |
                   S3C2410_LCDCON2_VSPW(3);

    lcd->LCDCON3 = S3C2410_LCDCON3_HBPD(19) |
                   S3C2410_LCDCON3_HOZVAL(HOZVAL_TFT) |
                   S3C2410_LCDCON3_HFPD(7);

    lcd->LCDCON4 = S3C2410_LCDCON4_MVAL(0) |
                   S3C2410_LCDCON4_HSPW(3);
    
    lcd->LCDCON5 = S3C2410_LCDCON5_FRM565 |
                   S3C2410_LCDCON5_INVVLINE |
                   S3C2410_LCDCON5_INVVFRAME |
                   S3C2410_LCDCON5_BSWP;

    lcd->LCDSADDR1 = (((u32)LCD_VIDEO_ADDR >> 22) << 21)|M5D((u32)LCD_VIDEO_ADDR >> 1);
    lcd->LCDSADDR2 = M5D(((u32)LCD_VIDEO_ADDR + (SCR_XSIZE_TFT * LCD_YSIZE_TFT * pGD->gdfBytesPP/1)) >> 1);
    lcd->LCDSADDR3 = (((SCR_XSIZE_TFT-LCD_XSIZE_TFT) / 2) << 11)|(LCD_XSIZE_TFT / 2);
    lcd->LCDINTMSK |= (3); // MASK LCD Sub Interrupt
    lcd->LPCSEL    &= (~7); // Disable LPC3600
    lcd->TPAL       = 0; // Disable Temp Palette
    //lcd->LCDINTPND=3;

    //============Lcd PowerEnable===========//
    //GPG4 is setted as LCD_PWREN
    gpio->GPGUP=gpio->GPGUP   & (~(1<<4))|(1<<4);  // Pull-up disable
    gpio->GPGCON=gpio->GPGCON & (~(3<<8))|(3<<8);  //GPG4=LCD_PWREN
    
    //Enable LCD POWER ENABLE Function
    lcd->LCDCON5 = lcd->LCDCON5 & (~(1<<3))|(1<<3);   // PWREN
    lcd->LCDCON5  = lcd->LCDCON5 &(~(1<<5))|(0<<5);   // INVPWREN
    
    /* Clear video memory */
    memset(LCD_VIDEO_ADDR, 0, CFG_VIDEO_LOGO_MAX_SIZE);
    
    // ENVID=ON,enable the LCD output
    lcd->LCDCON1 |= 1;


#ifdef CONFIG_SPLASH_SCREEN
    bmp_log_init(0);
#endif /* CONFIG_SPLASH_SCREEN */


}


/* The sum of all part_size[]s must equal to the NAND size, i.e., 0x4000000.
   "initrd" is sized such that it can hold two uncompressed 16 bit 640*480
   images: 640*480*2*2 = 1228800 < 1245184. */

unsigned int dynpart_size[] = {
    CFG_UBOOT_SIZE, 0x20000, 0x200000, 0xa0000, 0x3d5c000-CFG_UBOOT_SIZE, 0 };
char *dynpart_names[] = {
    "u-boot", "u-boot_env", "kernel", "splash", "rootfs", NULL };


